;;;-*- Mode: Lisp; Package: CCL -*-
;;;
;;;   Copyright (C) 1994-2001 Digitool, Inc
;;;   This file is part of OpenMCL.  
;;;
;;;   OpenMCL is licensed under the terms of the Lisp Lesser GNU Public
;;;   License , known as the LLGPL and distributed with OpenMCL as the
;;;   file "LICENSE".  The LLGPL consists of a preamble and the LGPL,
;;;   which is distributed with OpenMCL as the file "LGPL".  Where these
;;;   conflict, the preamble takes precedence.  
;;;
;;;   OpenMCL is referenced in the preamble as the "LIBRARY."
;;;
;;;   The LLGPL is also available online at
;;;   http://opensource.franz.com/preamble.html

(in-package "CCL")

;;; No error checking, no interrupts, no protect_caller, no nuthin.
;;; No error, no cons.  No problem.
(defun %progvrestore (saved)
  (declare (optimize (speed 3) (safety 0)))
  (dolist (pair saved)
    (%set-sym-value (car pair) (cdr pair))))

;;; Check that something that's supposed to be a proper list of
;;; symbols is; error otherwise.
;;; This is called only by the compiler output of a PROGV form.
;;; It checks for the maximum length that the progvsave subprim
;;; can handle.
(defun svar-check-symbol-list (l &optional (max-length
                                        (floor (- 4096 20) (* target::node-size 3))
                                       ))
  (let ((len (list-length l)))
    (if (and len
             (or (null max-length)
                 (< len max-length))
             (dolist (s l t) 
               (unless (and (symbolp s)
                            (not (constant-symbol-p s))
                            (not (logbitp $sym_vbit_global (the fixnum (%symbol-bits s)))))
                 (return nil))))
      (mapcar #'ensure-svar l)
      (error "~s is not a proper list of bindable symbols~@[ of length < ~s~]." l max-length))))

(defun check-symbol-list (l &optional (max-length
                                        (floor (- 4096 20) (* target::node-size 3))
                                       ))
  (let ((len (list-length l)))
    (if (and len
             (or (null max-length)
                 (< len max-length))
             (dolist (s l t) 
               (unless (and (symbolp s)
                            (not (constant-symbol-p s))
                            (not (logbitp $sym_vbit_global (the fixnum (%symbol-bits s)))))
                 (return nil))))
      l
      (error "~s is not a proper list of bindable symbols~@[ of length < ~s~]." l max-length))))

;;; The type-checking done on the "plist" arg shouldn't be removed.
(defun set-symbol-plist (sym plist)
  (let* ((pp (%symbol-package-plist sym))
         (newpp pp))
    (let* ((len (list-length plist)))
      (unless (and len (evenp len))
        (error "Bad plist: ~s" plist)))
    (if (atom pp)
      (setq newpp (cons pp plist))
      (let* ((bits (%symbol-bits sym)))
        (declare (fixnum bits))
        (if (logbitp $sym_vbit_typeppred bits)
          (setf (cdr (cdr pp)) plist)
          (setf (cdr pp) plist))))
    (%set-symbol-package-plist sym newpp)
    plist))


(defun %pl-search (l key)
  (declare (list l) (optimize (speed 3)))
  (loop
    (if (eq (car l) key)
      (return)
      (if l
        (setq l (cdr (the list (cdr l))))
        (return))))
  l)


(defun symbol-plist (sym)
  "Return SYMBOL's property list."
  (let* ((pp (%symbol-package-plist sym)))
    (if (consp pp)
      (let* ((bits (%symbol-bits sym)))
        (declare (fixnum bits))
        (if (logbitp $sym_vbit_typeppred bits)
          (cddr pp)
          (cdr pp))))))


(defun get (sym key &optional default)
  "Look on the property list of SYMBOL for the specified INDICATOR. If this
  is found, return the associated value, else return DEFAULT."
  (let* ((tail (%pl-search (symbol-plist sym) key)))
    (if tail (%cadr tail) default)))

(defun put (sym key value)
  (let* ((plist (symbol-plist sym))
         (tail (%pl-search plist key)))
    (if tail 
      (%rplaca (%cdr tail) value)
      (set-symbol-plist sym (cons key (cons value plist))))
    value))

;;; This is how things have traditionally been defined: if %sym_vbit_typeppred 
;;; is set in the symbol's %SYMBOL-BITS, it's assumed that it's safe to
;;; take the %CADR of the %SYMBOL-PACKAGE-PLIST.
(defun get-type-predicate (name)
  (let* ((bits (%symbol-bits name)))
    (declare (fixnum bits))
    (if (logbitp $sym_vbit_typeppred bits)
      (%cadr (%symbol-package-plist name)))))

(defun set-type-predicate (name function)
  (let* ((bits (%symbol-bits name))
         (spp (%symbol-package-plist name)))
    (declare (fixnum bits))
    (if (logbitp $sym_vbit_typeppred bits)
      (%rplaca (%cdr (%symbol-package-plist name)) function)
      (progn
        (%symbol-bits name (the fixnum (bitset $sym_vbit_typeppred bits)))
        (if (atom spp)
          (%set-symbol-package-plist name (cons spp (cons function nil)))
          (%rplacd spp (cons function (cdr spp))))))
    function))

(defun symbol-value (sym)
  "Return SYMBOL's current bound value."
  (let* ((val (%sym-value sym)))
    (if (eq val (%unbound-marker))
      (%kernel-restart $xvunbnd sym)
      val)))

(defun set (sym value)
  "Set SYMBOL's value cell to NEW-VALUE."
  (let* ((bits (%symbol-bits sym)))
    (declare (fixnum bits))
    (if (logbitp $sym_vbit_const bits)
      (%err-disp $XCONST sym)
      (%set-sym-value sym value))))

(defun constant-symbol-p (sym)
  (and (symbolp sym)
       (%ilogbitp $sym_vbit_const (%symbol-bits sym))))

;;; This leaves the SPECIAL bit alone, clears the others.
(defun makunbound (sym)
  "Make SYMBOL unbound, removing any value it may currently have."
  (if (and *warn-if-redefine-kernel*
           (constant-symbol-p sym))
    (cerror "Make ~S be unbound anyway."
            "~S is a constant; making it unbound might be a bad idea." sym))
  (%symbol-bits sym (the fixnum (logand (logior #xff00 (ash 1 $sym_bit_special))
                                        (the fixnum (%symbol-bits sym)))))
  (%set-sym-value sym (%unbound-marker))
  sym)

(defun non-nil-symbolp (x)
  "Returns symbol if true"
  (if (symbolp x) x))

(defun symbol-package (sym)
  "Return the package SYMBOL was interned in, or NIL if none."
  (let* ((pp (%symbol-package-plist sym)))
    (if (consp pp) (car pp) pp)))

(defun boundp (sym)
  "Return non-NIL if SYMBOL is bound to a value."
  (not (eq (%sym-value sym) (%unbound-marker))))

(defun make-symbol (name)
  "Make and return a new symbol with the NAME as its print name."
  (%gvector target::subtag-symbol
                (ensure-simple-string name) ; pname
                (%unbound-marker)       ; value cell
                %unbound-function%      ; function cell
                nil                     ; package&plist
                0))                     ; flags

(defun %symbol-bits (sym &optional new)
  (let* ((p (%symbol->symptr sym))
         (bits (%svref p target::symbol.flags-cell)))
    (if new
      (setf (%svref p target::symbol.flags-cell) new))
    bits))

(defun %sym-value (name)
  (let* ((symptr (%symbol->symptr name))
         (bits (%svref symptr target::symbol.flags-cell))
         (svar (unless (or (logbitp $sym_vbit_global bits)
                           (logbitp $sym_vbit_const bits))
                 (%find-svar symptr))))
    (declare (fixnum bits))
    (if svar
      (%svar-sym-value svar)
      (%svref symptr target::symbol.vcell-cell))))

(defun %set-sym-value (name val)
  (let* ((symptr (%symbol->symptr name))
         (bits (%svref symptr target::symbol.flags-cell))
         (svar (unless (or (logbitp $sym_vbit_global bits)
                           (logbitp $sym_vbit_const bits))
                 (%find-svar symptr))))
    (declare (fixnum bits))
    (if svar
      (%svar-set-sym-value svar val)
      (setf (%svref symptr target::symbol.vcell-cell) val))))
    
(defun %sym-global-value (name)
  (%svref (%symbol->symptr name) target::symbol.vcell-cell))

(defun %set-sym-global-value (name val)
  (setf (%svref (%symbol->symptr name) target::symbol.vcell-cell) val))

(defun %symbol-package-plist (sym)
  (%svref (%symbol->symptr sym) target::symbol.package-plist-cell))

(defun %set-symbol-package-plist (sym new)
  (setf (%svref (%symbol->symptr sym) target::symbol.package-plist-cell) new))



(defun symbol-name (sym)
  "Return SYMBOL's name as a string."
  #+ppc32-target
  (%svref (%symbol->symptr sym) ppc32::symbol.pname-cell)
  #+ppc64-target
  (if sym                               ;NIL's pname is implicit
    (%svref (%symbol->symptr sym) ppc64::symbol.pname-cell)
    "NIL")
  )




(defun %global-macro-function (symbol)
  (let* ((fbinding (fboundp symbol)))
    (if (and (typep fbinding 'simple-vector)
             (= (the fixnum (uvsize fbinding)) 2))
      (let* ((fun (%svref fbinding 1)))
        (if (functionp fun) fun)))))

(defun %symbol-binding-address (sym)
  (%svar-binding-address (ensure-svar sym)))

;;(defvar *interrupt-level* -1)

(let* ((svar-lock (make-lock))
       (svar-hash (make-hash-table :test #'eq :weak t))
       (svar-idx-map (make-hash-table :test #'eq :weak :value))
       (fixed-svar-symbols '())
       (svar-index (length fixed-svar-symbols)))
  (defun %set-svar-hash (hash)
    (unless svar-hash
      (setq svar-hash hash)))
  (defun %svar-hash ()
    svar-hash)
  (defun %find-svar (symptr)
    (gethash symptr svar-hash))
  (defun find-svar (sym)
    (%find-svar (%symbol->symptr sym)))
  (defun ensure-svar-idx (svar)
    (let* ((sym (%svref svar target::svar.symbol-cell))
           (idx (%svref svar target::svar.idx-cell))
           (bits (%symbol-bits sym)))
      (declare (fixnum idx bits))
      (if (or (logbitp $sym_vbit_global bits)
              (logbitp $sym_vbit_const bits))
        (unless (zerop idx)
	  (remhash idx svar-idx-map)
          (setf (%svref svar target::svar.idx-cell) 0))
        (if (zerop idx)
	  (let* ((new-idx
                  (or (do* ((i 1 (1+ i))
                            (fixed-syms fixed-svar-symbols (cdr fixed-syms)))
                           ((null fixed-syms))
                        (declare (fixnum i))
                        (when (eq (car fixed-syms) sym)
                          (return i)))
                      (incf svar-index))))
	    (setf (%svref svar target::svar.idx-cell) new-idx)
	    (setf (gethash new-idx svar-idx-map) svar))))
      svar))
  (defun %ensure-svar (symptr)
    (with-lock-grabbed (svar-lock)
      (let* ((found (%find-svar symptr))
             (svar (or found
                       (setf (gethash symptr svar-hash)
                             (gvector :svar symptr 0)))))
        (ensure-svar-idx svar))))
  (defun ensure-svar (sym)
    (%ensure-svar (%symbol->symptr sym)))
  (defun cold-load-svar (svar)
    (with-lock-grabbed (svar-lock)
      (let* ((symptr (%symbol->symptr (%svref svar target::svar.symbol-cell))))
        (setf (gethash symptr svar-hash) svar)
        (setf (%svref svar target::svar.idx-cell) 0)
        (ensure-svar-idx svar))))
  (defun svar-index ()
    svar-index)
  (defun index-svar (index)
    (gethash index svar-idx-map))
  )
